/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in 
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2007.  All Rights Reserved.
 *
 * Protege was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu.
 *
 */

package edu.stanford.smi.protege.server.framestore;

import java.lang.reflect.Method;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import edu.stanford.smi.protege.exception.OntologyException;
import edu.stanford.smi.protege.exception.ProtegeError;
import edu.stanford.smi.protege.exception.ProtegeIOException;
import edu.stanford.smi.protege.model.Frame;
import edu.stanford.smi.protege.model.framestore.AbstractFrameStoreInvocationHandler;
import edu.stanford.smi.protege.model.query.Query;
import edu.stanford.smi.protege.model.query.QueryCallback;
import edu.stanford.smi.protege.server.RemoteSession;
import edu.stanford.smi.protege.server.metaproject.MetaProject;
import edu.stanford.smi.protege.server.metaproject.ProjectInstance;
import edu.stanford.smi.protege.server.metaproject.User;

public class LastUsageInvocationHandler extends AbstractFrameStoreInvocationHandler {
    public static long ACCESS_TIME_GRANULARITY = 15 * 1000;
    private ProjectInstance projectInstance;
    private MetaProject metaproject;
    private Map<User, Date> lastAccessTimeMap = new HashMap<User, Date>();
    
    public LastUsageInvocationHandler(ProjectInstance projectInstance) {
        this.projectInstance = projectInstance;
        metaproject = projectInstance.getMetaProject();
    }

    @Override
    protected void executeQuery(Query q, final QueryCallback qc) {
        updateLastAccessTime();
        QueryCallback outerCallback = new QueryCallback() {
            
            public void provideQueryResults(Collection<Frame> results) {
                updateLastAccessTime();
                qc.provideQueryResults(results);
            }
            
            public void handleError(ProtegeError error) {
                updateLastAccessTime();
                qc.handleError(error);
            }
            
            public void handleError(ProtegeIOException exception) {
                updateLastAccessTime();
                qc.handleError(exception);
            }
            
            public void handleError(OntologyException exception) {
                updateLastAccessTime();
                qc.handleError(exception);
            }
        };
        getDelegate().executeQuery(q, outerCallback);
    }

    @Override
    protected Object handleInvoke(Method method, Object[] args) {
        if (!method.getName().equals("getEvents")) {
            updateLastAccessTime();
        }
        return invoke(method, args);
    }
    
    private void updateLastAccessTime() {
        RemoteSession session = ServerFrameStore.getCurrentSession();
        if (session == null) {
            return;
        }
        User u = metaproject.getUser(session.getUserName());
        Date now = new Date();
        Date then = lastAccessTimeMap.get(u);
        if (then == null || now.getTime() > then.getTime() + ACCESS_TIME_GRANULARITY) {
            u.setLastAccess(now);
            lastAccessTimeMap.put(u, now);
        }
    }

}
